#include <cstdio>
#include <stdlib.h>

#include "RakPeerInterface.h"
#include "RakVoice/RakVoiceInterface.h"
#include "RakNetworkFactory.h"
#include "RakVoice/RakVoiceFactory.h"
#include <conio.h>
#include <assert.h>
#include <memory.h>
#include <cstring>
#include "portaudio.h"
#include "RakNetStatistics.h"

RakVoiceInterface* rakVoice;
RakPeerInterface *rakPeer;

#define SAMPLE_RATE          (8000)
// Port audio defines this as number of input channels.  I'm not sure what that means but I think leaving it at one is best
#define SAMPLES_PER_FRAME        (1)
const int MAX_PLAYERS=4;

bool muted;
// Just so we can remember where the packet came from
bool isServer;
int framesPerBuffer;
int blockSize;

static int PACallback( void *inputBuffer, void *outputBuffer,
						  unsigned long framesPerBuffer,
						  PaTimestamp outTime, void *userData )
{
	static int i;
	static PlayerID sender;
	static PlayerID remoteSystems[8];
	unsigned short numberOfSystems;

	//assert(BLOCK_SIZE==framesPerBuffer);

#ifdef TEST_LOOPBACK
	// For testing - send to ourselves
	if (inputBuffer)
		rakVoice->EncodeSoundPacket((char*)inputBuffer,rakPeer->GetInternalID());
#endif

	if (!muted)
	{
		numberOfSystems=8;
		rakPeer->GetConnectionList(remoteSystems, &numberOfSystems);

		for (i=0; i < numberOfSystems; i++)
			// Since we are broadcasting, and EncodeSoundPacket has to specify a particular target, loop through everyone we know about.
			rakVoice->EncodeSoundPacket((char*)inputBuffer,remoteSystems[i]);
	}

	// Writes to outputBuffer if a sound packet is waiting.  Otherwise returns false and does nothing
	// Not using sender here, but if we had the interface for it we could use it show who is speaking
	if (rakVoice->GetSoundPacket((char*)outputBuffer, &sender)==false)
	{
		// If we didn't fill in anything, fill in silence
		memset(outputBuffer, 0, blockSize);
	}

	return 0;
}

#include "Multiplayer.h"
template <class InterfaceType>
class RakVoiceMultiplayer : public Multiplayer<InterfaceType>
{
public:
	void ReceiveVoicePacket(Packet *packet,InterfaceType *interfaceType);

	void ShowDialogText(char *str);
private:
};

template <class InterfaceType>
void RakVoiceMultiplayer<InterfaceType>::ReceiveVoicePacket(Packet *packet,InterfaceType *interfaceType)
{
	// Put a packet in the queue
	rakVoice->DecodeAndQueueSoundPacket((char*)packet->data, packet->length);
}

template <class InterfaceType>
void RakVoiceMultiplayer<InterfaceType>::ShowDialogText(char *str)
{
	printf("%s",str);
}


/*******************************************************************/
int main(void);
int main(void)
{
	int i;
	char ch;
	RakVoiceMultiplayer<RakPeerInterface> rakVoiceMultiplayer;
	rakPeer = RakNetworkFactory::GetRakPeerInterface();
	rakVoice = RakVoiceFactory::GetRakVoiceInterface();

	// Crude interface
	// Holds user data
	char ip[30], serverPort[30], clientPort[30];
	bool b;
	char c;

	// A client
	isServer=false;

	// Get our input
	puts("Enter the local listening port");
	gets(clientPort);
	if (clientPort[0]==0)
		strcpy(clientPort, "60000");

	
	b=rakPeer->Initialize(8, atoi(clientPort), 30);

	if (b)
		puts("Starting RakPeer");
	else
	{
		puts("Failed to start RakPeer.  Terminating.");
		exit(1);
	}

	rakPeer->SetMaximumIncomingConnections(8);
	

	int bitsPerSample = 8;
	int sampleType;
	int numberOfBlocks;

	// 16 bits per sample doesn't really work right now.  I don't know why.
	if (bitsPerSample==16)
	{
        numberOfBlocks=10;
		sampleType=paInt16;
	}
	else
	{
		numberOfBlocks=10;
		sampleType=paInt8;
	}

	rakVoice->Init(SAMPLE_RATE, bitsPerSample, rakPeer);
	framesPerBuffer=numberOfBlocks*rakVoice->GetFrameSize();
	blockSize=SAMPLES_PER_FRAME * framesPerBuffer;

	// A block is an array of data of length blockSize.  One block is passed to PACallback and is encoded to speex
	rakVoice->SetBlockSize(blockSize);

	PaError    err;
	PortAudioStream *stream;
	err = Pa_Initialize();
	if( err != paNoError ) goto error;
	muted=false;

	/* Record some audio. -------------------------------------------- */
	err = Pa_OpenStream(
		&stream,
		Pa_GetDefaultInputDeviceID(),
		SAMPLES_PER_FRAME,
		sampleType,
		NULL,
		Pa_GetDefaultOutputDeviceID(),
		SAMPLES_PER_FRAME,
		sampleType,
		NULL,
		SAMPLE_RATE,
		framesPerBuffer,            /* frames per buffer */
		0,               /* number of buffers, if zero then use default minimum */
		0, /* paDitherOff, // flags */
		PACallback,
		0);
	if( err != paNoError ) goto error;

	err = Pa_StartStream( stream );
	if( err != paNoError ) goto error;
	printf("Now recording!!\n"); fflush(stdout);

	puts("Hit 'm' to toggle mute (default off).\nHit 'c' to connect to another system.\nHit 's' to show network statistics.\nHit 'p' to get the connected player list.\nHit 'q' to quit.");
	ch=0;
	while( ch!='q')
	{
		if (kbhit())
		{
			ch=getch();
			if (ch=='m')
			{
				muted=!muted;
				if (muted)
					puts("Microphone muted.");
				else
					puts("Microphone no longer muted.");
				ch=0;
			}
			else if (ch=='c')
			{
				puts("Enter IP to connect to");
				gets(ip);
				if (ip[0]==0)
					strcpy(ip, "127.0.0.1");
				puts("Enter the port to connect to");
				gets(serverPort);
				if (serverPort[0]==0)
					strcpy(serverPort, "60000");
				// Connecting the client is very simple.  0 means we don't care about
				// a connectionValidationInteger, and false for low priority threads
				bool b = rakPeer->Connect(ip, atoi(serverPort),  0, 0);
				if (b==false)
					puts("Connection call failed");
				else
					puts("Connection call complete");

				c=0;
			}
			else if (ch=='p')
			{
				PlayerID remoteSystems[8];
				unsigned short numberOfSystems;

				numberOfSystems=8;
				rakPeer->GetConnectionList(remoteSystems, &numberOfSystems);

				if (numberOfSystems>0)
				{
					printf("Connected players:\n");
					for (i=0; i < numberOfSystems; i++)
						printf("%i. %s:%i\n", i+1, rakPeer->PlayerIDToDottedIP(remoteSystems[i]), remoteSystems[i].port);
				}
				else
					printf("Nobody connected.\n");				

				ch=0;
			}
			else if (ch=='s')
			{
				char buffer[2048];		
				StatisticsToString(rakPeer->GetStatistics(rakPeer->GetPlayerIDFromIndex(0)),buffer,1);
				printf("%s",buffer);
				ch=0;
			}
		}
		rakVoiceMultiplayer.ProcessPackets(rakPeer);
		Pa_Sleep(30);
	}

	err = Pa_CloseStream( stream );
	if( err != paNoError ) goto error;

	rakVoice->Deinit();
	RakVoiceFactory::DestroyRakVoiceInterface(rakVoice);
	RakNetworkFactory::DestroyRakPeerInterface(rakPeer);

	printf("Full duplex sound test complete.\n" );
	fflush(stdout);
	return 0;

error:
	Pa_Terminate();
	fprintf( stderr, "An error occurred while using the portaudio stream\n" );
	fprintf( stderr, "Error number: %d\n", err );
	fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
	return -1;
}
